/*
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010.
 * See http://www.eu-egee.org/partners/ for details on the
 * copyright holders. 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.glite.apel.condor;



import org.apache.log4j.Logger;

import org.glite.apel.core.Log;

import org.glite.apel.core.Time;

import org.glite.apel.core.TimeInterval;

import org.glite.apel.core.log.type.EventRecord;

import org.glite.apel.core.log.type.EventRecordIterator;



import java.io.BufferedReader;

import java.io.File;

import java.io.FileInputStream;

import java.io.IOException;

import java.io.InputStreamReader;

import java.text.SimpleDateFormat;

import java.util.Calendar;

import java.util.Date;

import java.util.TimeZone;

import java.util.Vector;

import java.util.regex.Pattern;

import java.util.zip.GZIPInputStream;

/**

 * A file containing a CONDOR event log

 */

class EventLogParser extends File {   

    /**

     * Using log4j logging tool.

     */

    protected static final Logger LOG = 

        Logger.getLogger(EventLogParser.class.getName());

    /**

     * Buffer which reads-in data from the log file.

     */

    protected BufferedReader m_reader;

    /**

     * Holds all event records read-in from the event log file.

     */

    protected Vector m_eventRecords;

    /** 

     * The format used to encode the current date. 

     */

    private static final String EVENT_DATE_FORMAT = "yyyyMMdd";

    /** 

     * UTC constant used when converting the current date. 

     */

    private static final String UTC = "UTC";

    /**

     * Formatter converts the current date into the same

     * format used by the event log file.

     */

    private SimpleDateFormat m_dateFormatter;

    /*
     * A file object representing a single particular log file
     */
    private File m_file;

    /**

     * The timezone to be used when interpreting time stamps 

     * parsed from the log files.

     */

    private String m_timezone;

    /**

     * Pattern used to chop up a line of data that matches an

     * EventRecord into a convenient list.

     */

    private Pattern m_delimiter;

    /**

     * Pattern used to match a condor event record in the log file.

     */

    private Pattern m_endEventPattern;

    /**

     * Pattern used to match a condor event record in the log file.

     */

    private Pattern m_startEventPattern;

    /**

     * Patterns used to identify key elements within the list 

     */

    private Pattern pattern=Pattern.compile("\\s=\\s");

    private Pattern tmpA_pattern;

    private Pattern tmpB_pattern;

    /**

     * Condor logs don't report the batch submit host: This is added by means of a

     * <SubmitHost>value</SumitHost> token in APEL Event log config file.

     */

     private String m_submitHost;

    /**

     * Creates a new event log file.

     */

    EventLogParser(File file, String timezone, String submitHost) {

        super(file.getPath());

        m_eventRecords = new Vector();

        m_submitHost = submitHost;

        m_file = file;

        m_timezone = timezone;

        //m_dateFormatter = new SimpleDateFormat(EVENT_DATE_FORMAT);

        //UTC conversion is wrong! 

        //m_dateFormatter.setTimeZone(TimeZone.getTimeZone(UTC));

        m_delimiter = Pattern.compile(";;;");

        //String prefixString = "^\\*{3}$";

        String prefixString = "^\\*{3}.*$";

        m_endEventPattern = Pattern.compile(prefixString);

        String fixString = "^MyType = \"Job\"$";

        m_startEventPattern = Pattern.compile(fixString);

    }


    public void openLogFile() {
        try {

            FileInputStream fileInputStream = new FileInputStream(m_file);

            if (getName().matches("^.*\\.gz$")) {

                m_reader = new BufferedReader(new InputStreamReader(

                    new GZIPInputStream(fileInputStream)));

            } else {

                m_reader = new BufferedReader(new InputStreamReader(

                    fileInputStream));

            }

        } catch (Exception e) {

            Log.warn(LOG, "Exception opening file " + m_file.getPath(), e);

        }
    }

    /**

     * Returns true if the file name matches today's date (ie the log

     * is currently being used).

     * ===>

     * ===> What happens if the log does not have a date and is always active?

     * ===>

     */

    boolean isActiveLog() {

    /*

     *  Returns True is log is curently active

     *  Support history.YYYYMMDD.gz format for Cambridge using APEL log rortaion

                and history.1-365.gz for OSG

     *  To Do: Support condor 6.7.x with imnplict log rotation:

     *         See Savannah Bug report # 16657

     */

        SimpleDateFormat m_dateFormatter = new SimpleDateFormat(EVENT_DATE_FORMAT);

        m_dateFormatter.setTimeZone(TimeZone.getDefault());

        Date currentDate = Calendar.getInstance().getTime();

        String formattedDate = m_dateFormatter.format(currentDate);

        String EventLogPrefix = "history.";

        String EventLogPostfix = ".gz";

        String CurrentLog = EventLogPrefix + formattedDate + EventLogPostfix;

        Log.info(LOG,"CONDOR: CurrentLog = " + CurrentLog);

        return CurrentLog.equals(getName());

        //return true; 

    }

    /**

     * Parses the log file and returns an iterator over all the 

     * event records found within the file.

     */

    public void parse() {

        Log.info(LOG, "Processing " + getFullPath());

        String line = null;

        try {

        	/**

        	 * Condor lines are ClassAd key=value pairs. 

                 * The job record begins with line MyType=Job, and ends with the line "***"

        	 * Therefore, it is necessary to construct a record as a list of classAds seperated 

                 * by an arbitrary quantifier taken to be ;;;

        	 * Thus 

        	 *      MyType=Job

        	 *      Key1=Val1

        	 *      Key2=Val2

        	 *      ***

        	 * becomes

        	 *      MyType=Job;;;Key1=Val1;;;Key2=Val2;;;

        	 */

            String JobRecord=""; 

            while ((line = m_reader.readLine()) != null) {

                if (m_startEventPattern.matcher(line).matches()) {

                   // reset JobRecord

                   //System.out.println("Found start of JobRecord (MyType = \"Job\")");                   

                   JobRecord = "";

                } else if (m_endEventPattern.matcher(line).matches()) {

                        //System.out.println("Found end of JobRecord (***)");

                	EventRecord record = createEventRecord(JobRecord);

                        // add record to the list only if parameters defined

                        String tmp1=record.getJobName();

                        String tmp2=record.getSubmitHost();

                        if (tmp1 != null && tmp2 != null) {

                          //System.out.println("Add Record: JobName=" + tmp1 + " SubmitHost=" + tmp2 );

                          m_eventRecords.add(record);

                        }        

                } else {

                        // construct JobRecord

                	JobRecord = JobRecord.concat(line) + ";;;";

                }

            } // end while

        } catch (IOException e) {

            Log.warn(LOG, "Exception reading file " + getName(), e);

        }

    } // end method

    /**

     * Returns a new EventRecord derived from the line of data 

     * (parsed from the event log). 

     */

    EventRecord createEventRecord(String line) {

        EventRecord record = new EventRecord();

        // fields will be ClassAds e.g. fields[0] will be MyType=Job

        String[] fields = m_delimiter.split(line);

        //

        // Initialisation step for each record

        //

        // submitHost is determined from the User ClassAd

        record.setSubmitHost("User-ClassAd-NotFound");

        record.setSubmitHost(m_submitHost);

        Double remoteSysCpuTime = new Double(0);

        Double remoteUsrCpuTime = new Double(0);

        Double localSysCpuTime  = new Double(0);

        Double localUsrCpuTime  = new Double(0); 

        // When both are true, the total system CPU can be determined

        boolean rsc=false; // true if RemoteSystemCpu has been set

        boolean ruc=false; // true if RemoteUserCpu   has been set

        //

        // End Initialisation step

        //

       for (int i = 0; i < fields.length; i ++ ){

         	//System.out.println("fields[" + i + "] => " + fields[i]);

         	String[] ClassAds = pattern.split(fields[i]);

         // Split the ClassAds into Key=Value pairs

         for (int j = 0; j < ClassAds.length; j+=2 ){

       	  //System.out.println(ClassAds[j] + "=" + ClassAds[j+1] );

          ////////////////////////////////////////////////////////////////////////////

       	  // Local Batch JobID = jobNumber (= ClusterID e.g. 6)

       	  // jobNumber is later over-written if LastRemoteHost is found (6.host.mydomain)

          ////////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("ClusterId")) {

       	   String jobNumber=(String)ClassAds[j+1];

       	   record.setJobName(jobNumber);

       	   //System.out.println("===> JobName=" + jobNumber);

       	  }

          ////////////////////////////////////////////////////////////////////////////

       	  // Local User  (= Owner e.g. lassAds[j].equals("Owner")) {

       	  // Local Group (= Owner with trailing digits stripped e.g dteam)

          ////////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("Owner")) {

       	   String owner=(String)ClassAds[j+1].replaceAll("\"","");

           record.setLocalUserID(owner);

       	   //System.out.println("===> LocalUserID=" + owner);

       	   String group=(String)ClassAds[j+1].replaceAll("[\"\\d{3}]","");

       	   //System.out.println("===> LocalUserGroup=" + group);

           record.setLocalUserGroup(group);

       	  }

          /////////////////////////////////////////////////////////////////////////////

          // Get SubmitHost from User class ad?

          // User = "markc@serv03.hep.phy.cam.ac.uk"

          ///////////////////////////////////////////////////////////////////////////// 

          //if (ClassAds[j].equals("User")) {

          // String tmp=(String)ClassAds[j+1].replaceAll("\"","");

          // String[] tmpf = Pattern.compile("\\@").split(tmp);

          // record.setSubmitHost(tmpf[1]);

          // System.out.println("====> User_0:" + tmpf[0] + "====> User_1:" + tmpf[1]);

          //}

       	  ////////////////////////////////////////////////////////////////////////////

       	  // EDG_WL_JOBID=https://egee-rb-01.cnaf.infn.it:9000/tWyRXWWiaJglFMUThBOjOQ

          ////////////////////////////////////////////////////////////////////////////

       	  String EDG_WL_JOBID="null";

       	  tmpA_pattern = Pattern.compile(";");

       	  tmpB_pattern = Pattern.compile("^EDG_WL_JOBID=" + ".*");

       	  if (ClassAds[j].equals("Env")) {

       		  //System.out.println("[1] EDG_WL_JOBID =>" + ClassAds[j] + " = " + ClassAds[j+1]);

           	  String[] tmp = tmpA_pattern.split(ClassAds[j+1]);

           	  for (int k = 0; k < tmp.length; k++ ){

           		  if (tmpB_pattern.matcher(tmp[k]).matches()) {

           			  EDG_WL_JOBID=tmp[k];

           		  	  //System.out.println("===> EDG_WL_JOBID=" + EDG_WL_JOBID);

           		  }

           	  }

          }

       	  ////////////////////////////////////////////////////////////////////////////

       	  // RemoteWallClockTime

       	  // The RemoteWallClock appears twice in each per job record. 

       	  // We assume that the second instance over-writes the first

       	  ////////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("RemoteWallClockTime")) {

       		 Double wctTime=new Double((String)ClassAds[j+1]);  // put wct into a Double

       	         double cpu = Math.round(wctTime.doubleValue());    // 

       	         wctTime=new Double(cpu);

       	         record.setWallDurationSeconds(Integer.toString(wctTime.intValue()));

                 //System.out.println("===> WallDurationSeconds=" + Integer.toString(wctTime.intValue()));           

       		 String wctString=secondsToString(wctTime.doubleValue());

       	         //System.out.println("==> RemoteWallClockTime: CpuDurationSeconds into hh::mm:ss = " + wctString);

       	         TimeInterval interval;

       	         interval = new TimeInterval(wctString);

       	         //System.out.println("===> WallDuration=" + interval.getIso8601());

       	         record.setWallDuration(interval.getIso8601());

          }

       	  ////////////////////////////////////////////////////////////////////////////

       	  // System CPU Time = RemoteSysCpu + RemoteUserCpu

       	  // Note: LocalSysCpu and LocalUserCpu are ignored because we believe that they

       	  // are relevant only to the submit node and not the worker node.

       	  ////////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("RemoteSysCpu")) {

              rsc=true;

              remoteSysCpuTime=new Double((String)ClassAds[j+1]);  // put system cpu into a Double

       	  }

       	  if (ClassAds[j].equals("RemoteUserCpu")) {

                  ruc=true;

                  remoteUsrCpuTime=new Double((String)ClassAds[j+1]);  // put system cpu into a Double

       	  }

       	  // process the cpu information

          // What would happen if one of these were missing?

       	  if (rsc && ruc) {

       		  ruc=false;rsc=false;

       	      Double cpuTime = new Double(remoteSysCpuTime.doubleValue()+remoteUsrCpuTime.doubleValue());

       	      double cpu_sys = Math.round(cpuTime.doubleValue()); // 

      	       cpuTime=new Double(cpu_sys);

      	       record.setCpuDurationSeconds(Integer.toString(cpuTime.intValue()));   

               //System.out.println("===> CpuDurationSeconds=" + Integer.toString(cpuTime.intValue())); 

               String cpu_string = secondsToString(cpuTime.doubleValue());

               //System.out.println("===> RemoteSystemTime: CpuDurationSeconds into hh::mm:ss = " + cpu_string);

               TimeInterval interval = new TimeInterval(cpu_string);

               //System.out.println("===> CpuDuration=" + interval.getIso8601());

               record.setCpuDuration(interval.getIso8601());

       	  }

       	  ////////////////////////////////////////////////////////////////////////////

       	  // TIMEZONE

       	  ////////////////////////////////////////////////////////////////////////////

       	  TimeZone convertedTimezone = null;

       	  TimeZone timezone = TimeZone.getTimeZone(m_timezone);

          ////////////////////////////////////////////////////////////////////////////

       	  // StartTime (= JobStartDate)

          ////////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("JobStartDate")) {

       	        Time time = new Time((String)ClassAds[j+1], timezone);

       	        record.setStartTimeEpoch(time.getEpoch());

       	        record.setStartTime(time.getIso8601());

       	        record.setStartTimeUtc(time.getIso8601Utc());

       	        //System.out.println("===> StartTimeEpoch = " + time.getEpoch());

       	        //System.out.println("===> StartTimeUtc   = " + time.getIso8601Utc());

       	        //System.out.println("===> StartTime      = " + time.getIso8601());      

       	  }

          ////////////////////////////////////////////////////////////////////////////

       	  // CompletionDate: There are two CompletionDate ClassAds per job record 

       	  // although its not clear if this is always the case.

       	  // StopTime is taken to be second instance

          ////////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("CompletionDate")) {

       	        Time time = new Time((String)ClassAds[j+1], timezone);

       	        record.setStopTimeEpoch(time.getEpoch());

       	        record.setStopTime(time.getIso8601());

       	        record.setStopTimeUtc(time.getIso8601Utc());

       	        record.setEventDate(time.getIso8601Utc().substring(0, 10));

       	        record.setEventTime(time.getIso8601Utc().substring(11, 19));

       	        //System.out.println("===> StopTimeEpoch = " + time.getEpoch());

       	        //System.out.println("===> StopTimeUtc =   " + time.getIso8601Utc());

       	        //System.out.println("===> StopTime  =     " + time.getIso8601()); 

                //System.out.println("===> EventDate =     " + time.getIso8601Utc().substring(0, 10));

                //System.out.println("===> EventTime =     " + time.getIso8601Utc().substring(11, 19));     

       	  }

       	  /////////////////////////////////////////////////////////////////////////////////////////////

       	  // LastRemoteHost: (not described in classAd documentation)

       	  /////////////////////////////////////////////////////////////////////////////////////////////

       	  //if (ClassAds[j].equals("LastRemoteHost")) {

       	  //	  String lrh=(String)ClassAds[j+1].replaceAll("\"","");

       	  //	  System.out.println("===> LastRemoteHost =" + lrh);   

       	  //}

       	  ////////////////////////////////////////////////////////////////////////

       	  // Memory: The amount of RAM units MB

       	  // Note: This ClassAd was not available in the use case history logs

       	  //       Cast value into Kbytes

       	  ////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("Memory")) {

       		  Double memReal = new Double((String)ClassAds[j+1]);

       		  Double memRealVal = new Double(1000*memReal.doubleValue());

       		  //System.out.println("===> memReal = " + memRealVal.doubleValue() + " KB");

                  //=> NOT TESTED: System.out.println("===> memReal = " + memRealVal.toString() + " KB");  

       		  record.setMemoryReal(memRealVal.toString());

                  //System.out.println("====> MemoryReal=" + memRealVal.toString());

       	  }

       	  ////////////////////////////////////////////////////////////////////////

       	  // VirtualMemory units Kbytes

       	  ////////////////////////////////////////////////////////////////////////

       	  if (ClassAds[j].equals("VirtualMemory")) {

       		  //System.out.println("===> virtReal = " + (String)ClassAds[j+1] + " KB");

       		  record.setMemoryVirtual((String)ClassAds[j+1]);

                  //System.out.println("====> MemoryVirtual=" + (String)ClassAds[j+1]);

       	  } 	  

         }// End loop over the ClassAd Key=Value Pair

       }// End Loop over ClassAd Pairs

        return record;

    }

    /**

     * Remove the everything from the start of the string to the 

     * specified character.

     *

     * @param string string to strip.

     * @param delimiter character to look for.

     * @return string with leading text stripped off.

     */

    private String strip(String string, char delimiter) {

        int index = string.indexOf(delimiter);

        if (index != -1) {

            return string.substring(index + 1);

        }

        return string;

    }

    /**

     * Closes down the log file object. This is called once all meaningful 

     * data has been read from the file.

     */

    public void close() {

        if (m_reader != null) {

            try {

                m_reader.close();

            } catch (IOException e) {

                Log.warn(LOG, "Exception closing file " + getName(), e);

            }

        }

    }

    /**

     * Return canonical path, with built-in exception handling.

     * If an exception is thrown only the relative path is returned.

     */

    public String getFullPath() {

        String path;

        try {

            path = getCanonicalPath();

        } catch (Exception e) {

            Log.warn(LOG, "Exception trying to get filename", e);

            path = getName();

        }

        return path;

    }

    /**

     * Returns an iterator of all the parsed event records.

     */

    public EventRecordIterator getRecords() {

        return new EventRecordIterator(m_eventRecords.iterator());

    }

    /**

     * Converts time in seconds to a <code>String</code> in the 

     * format HH:mm:ss.

     */

    private String secondsToString(double time) {

     int seconds = (int) (time % 60);

     int minutes = (int) ((time/60) % 60);

     int hours = (int) ((time/3600) % 24);

     int days = (int) ((time/3600)/24);

     hours += days * 24;

     String secondsStr = (seconds < 10 ? "0" : "") + seconds;

     String minutesStr = (minutes < 10 ? "0" : "") + minutes;

     String hoursStr = (hours < 10 ? "0" : "") + hours;

     return new String(hoursStr + ":" + minutesStr + ":" + secondsStr);

    }

}

